#ifndef DALI_TOOLKIT_INTERNAL_TEXT_FIELD_H
#define DALI_TOOLKIT_INTERNAL_TEXT_FIELD_H

/*
 * Copyright (c) 2021 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

// EXTERNAL INCLUDES
#include <dali/devel-api/adaptor-framework/clipboard-event-notifier.h>
#include <dali/devel-api/adaptor-framework/input-method-context.h>
#include <dali/devel-api/atspi-interfaces/editable-text.h>
#include <dali/devel-api/atspi-interfaces/hypertext.h>
#include <dali/devel-api/atspi-interfaces/text.h>

// INTERNAL INCLUDES
#include <dali-toolkit/devel-api/controls/control-devel.h>
#include <dali-toolkit/devel-api/controls/text-controls/text-field-devel.h>
#include <dali-toolkit/internal/controls/control/control-data-impl.h>
#include <dali-toolkit/internal/text/decorator/text-decorator.h>
#include <dali-toolkit/internal/text/rendering/text-renderer.h>
#include <dali-toolkit/internal/text/text-anchor-control-interface.h>
#include <dali-toolkit/internal/text/text-control-interface.h>
#include <dali-toolkit/internal/text/text-controller.h>
#include <dali-toolkit/internal/text/text-editable-control-interface.h>
#include <dali-toolkit/internal/text/text-selectable-control-interface.h>
#include <dali-toolkit/public-api/controls/control-impl.h>
#include <dali-toolkit/public-api/controls/text-controls/text-field.h>

namespace Dali
{
namespace Toolkit
{
namespace Internal
{
/**
 * @brief A control which renders a short text string.
 */
class TextField : public Control, public Text::ControlInterface, public Text::EditableControlInterface, public Text::SelectableControlInterface, public Text::AnchorControlInterface
{
public:
  /**
   * @copydoc Dali::Toollkit::TextField::New()
   */
  static Toolkit::TextField New();

  // Properties

  /**
   * @brief Called when a property of an object of this type is set.
   *
   * @param[in] object The object whose property is set.
   * @param[in] index The property index.
   * @param[in] value The new property value.
   */
  static void SetProperty(BaseObject* object, Property::Index index, const Property::Value& value);

  /**
   * @brief Called to retrieve a property of an object of this type.
   *
   * @param[in] object The object whose property is to be retrieved.
   * @param[in] index The property index.
   * @return The current value of the property.
   */
  static Property::Value GetProperty(BaseObject* object, Property::Index index);

  /**
   * @copydoc Dali::Toollkit::TextField::GetInputMethodContext()
   */
  InputMethodContext GetInputMethodContext();

  /**
   * Connects a callback function with the object's signals.
   * @param[in] object The object providing the signal.
   * @param[in] tracker Used to disconnect the signal.
   * @param[in] signalName The signal to connect to.
   * @param[in] functor A newly allocated FunctorDelegate.
   * @return True if the signal was connected.
   * @post If a signal was connected, ownership of functor was passed to CallbackBase. Otherwise the caller is responsible for deleting the unused functor.
   */
  static bool DoConnectSignal(BaseObject* object, ConnectionTrackerInterface* tracker, const std::string& signalName, FunctorDelegate* functor);

  /**
   * @brief Gets text controller
   *
   * @return The text controller
   */
  Text::ControllerPtr GetTextController();

  /**
   * @copydoc TextField::TextChangedSignal()
   */
  Toolkit::TextField::TextChangedSignalType& TextChangedSignal();

  /**
   * @copydoc TextField::MaxLengthReachedSignal()
   */
  Toolkit::TextField::MaxLengthReachedSignalType& MaxLengthReachedSignal();

  /**
   * @copydoc TextField::TextChangedSignal()
   */
  Toolkit::TextField::InputStyleChangedSignalType& InputStyleChangedSignal();

  /**
   * @copydoc TextField::AnchorClickedSignal()
   */
  DevelTextField::AnchorClickedSignalType& AnchorClickedSignal();

  /**
   * @copydoc TextField::CursorPositionChangedSignal()
   */
  DevelTextField::CursorPositionChangedSignalType& CursorPositionChangedSignal();

  /**
   * @copydoc TextField::InputFilteredSignal()
   */
  DevelTextField::InputFilteredSignalType& InputFilteredSignal();

  /**
   * @copydoc TextField::SelectionChangedSignal()
   */
  DevelTextField::SelectionChangedSignalType& SelectionChangedSignal();

  /**
   * @copydoc TextField::SelectionClearedSignal()
   */
  DevelTextField::SelectionClearedSignalType& SelectionClearedSignal();

private: // From Control
  /**
   * @copydoc Control::OnInitialize()
   */
  void OnInitialize() override;

  /**
   * @copydoc Control::OnStyleChange()
   */
  void OnStyleChange(Toolkit::StyleManager styleManager, StyleChange::Type change) override;

  /**
   * @copydoc Control::GetNaturalSize()
   */
  Vector3 GetNaturalSize() override;

  /**
   * @copydoc Control::GetHeightForWidth()
   */
  float GetHeightForWidth(float width) override;

  /**
   * @copydoc Control::OnInitialize()
   */
  void OnRelayout(const Vector2& size, RelayoutContainer& container) override;

  /**
   * @copydoc Control::OnKeyInputFocusGained()
   */
  void OnKeyInputFocusGained() override;

  /**
   * @copydoc Control::OnKeyInputFocusLost()
   */
  void OnKeyInputFocusLost() override;

  /**
   * @copydoc Control::OnAccessibilityActivated()
   */
  bool OnAccessibilityActivated() override;

  /**
   * @copydoc Control::OnTap()
   */
  void OnTap(const TapGesture& tap) override;

  /**
   * @copydoc Control::OnPan()
   */
  void OnPan(const PanGesture& gesture) override;

  /**
   * @copydoc Control::OnLongPress()
   */
  void OnLongPress(const LongPressGesture& gesture) override;

  /**
   * @copydoc Control::OnSceneConnection()
   */
  void OnSceneConnection(int depth) override;

  /**
   * @copydoc Dali::CustomActorImpl::OnKeyEvent(const KeyEvent&)
   */
  bool OnKeyEvent(const KeyEvent& event) override;

  // From ControlInterface

  /**
   * @copydoc Text::ControlInterface::RequestTextRelayout()
   */
  void RequestTextRelayout() override;

  // From EditableControlInterface

  /**
   * @copydoc Text::EditableControlInterface::TextChanged()
   */
  void TextInserted(unsigned int position, unsigned int length, const std::string& content) override;

  /**
   * @copydoc Text::EditableControlInterface::TextDeleted()
   */
  void TextDeleted(unsigned int position, unsigned int length, const std::string& content) override;

  /**
   * @copydoc Text::EditableControlInterface::CursorPositionChanged()
   */
  void CursorPositionChanged(unsigned int oldPosition, unsigned int newPosition) override;

  /**
   * @copydoc Text::EditableControlInterface::TextChanged()
   */
  void TextChanged(bool immediate) override;

  /**
   * @copydoc Text::EditableControlInterface::MaxLengthReached()
   */
  void MaxLengthReached() override;

  /**
   * @copydoc Text::EditableControlInterface::InputStyleChanged()
   */
  void InputStyleChanged(Text::InputStyle::Mask inputStyleMask) override;

  /**
   * @copydoc Text::SelectableControlInterface::SelectionChanged()
   */
  void SelectionChanged(uint32_t oldStart, uint32_t oldEnd, uint32_t newStart, uint32_t newEnd) override;

  /**
   * @copydoc Text::EditableControlInterface::AddDecoration()
   */
  void AddDecoration(Actor& actor, bool needsClipping) override;

  /**
   * @copydoc Text::EditableControlInterface::InputFiltered()
   */
  void InputFiltered(Toolkit::InputFilter::Property::Type type) override;

  /**
   * @copydoc Text::EditableControlInterface::GetControlBackgroundColor()
   */
  void GetControlBackgroundColor(Vector4& color) const override;

  // From SelectableControlInterface
public:
  /**
   * @copydoc Text::SelectableControlInterface::SetTextSelectionRange()
   */
  void SetTextSelectionRange(const uint32_t* start, const uint32_t* end) override;

  /**
   * @copydoc Text::SelectableControlInterface::GetTextSelectionRange()
   */
  Uint32Pair GetTextSelectionRange() const override;

  /**
   * @copydoc Text::SelectableControlInterface::SelectWholeText()
   */
  void SelectWholeText() override;

  /**
   * @copydoc Text::SelectableControlInterface::SelectNone()
   */
  void SelectNone() override;

  /**
   * @copydoc Text::SelectableControlInterface::SelectText()
   */
  void SelectText(const uint32_t start, const uint32_t end) override;

  /**
   * @copydoc Text::SelectableControlInterface::GetSelectedText()
   */
  string GetSelectedText() const override;

  /**
   * @copydoc Text::EditableControlInterface::IsEditable()
   */
  bool IsEditable() const override;

  /**
   * @copydoc Text::EditableControlInterface::SetEditable()
   */
  void SetEditable(bool editable) override;

  /**
   * @copydoc Dali::EditableControlInterface::CopyText()
   */
  string CopyText() override;

  /**
   * @copydoc Dali::EditableControlInterface::CutText()
   */
  string CutText() override;

  /**
   * @copydoc Text::EditableControlInterface::PasteText()
   */
  void PasteText() override;

  // From AnchorControlInterface

  /**
   * @copydoc Text::AnchorControlInterface::AnchorClicked()
   */
  void AnchorClicked(const std::string& href) override;

  /**
   * @brief Get the rendered size of a specific text range.
   * if the requested text is at multilines, multiple sizes will be returned for each text located in a separate line.
   * if a line contains characters with different directions, multiple sizes will be returned for each block of contiguous characters with the same direction.
   *
   * @param[in] startIndex start index of the text requested to calculate size for.
   * @param[in] endIndex end index(included) of the text requested to calculate size for.
   * @return list of sizes of the reuested text.
   */
  Vector<Vector2> GetTextSize(const uint32_t startIndex, const uint32_t endIndex) const;

  /**
   * @brief Get the top/left rendered position of a specific text range.
   * if the requested text is at multilines, multiple positions will be returned for each text located in a separate line.
   * if a line contains characters with different directions, multiple positions will be returned for each block of contiguous characters with the same direction.
   *
   * @param[in] startIndex start index of the text requested to get position to.
   * @param[in] endIndex end index(included) of the text requested to get position to.
   * @return list of positions of the requested text.
   */
  Vector<Vector2> GetTextPosition(const uint32_t startIndex, const uint32_t endIndex) const;

private: // Implementation
  /**
   * @copydoc Dali::Toolkit::Text::Controller::(InputMethodContext& inputMethodContext, const InputMethodContext::EventData& inputMethodContextEvent)
   */
  InputMethodContext::CallbackData OnInputMethodContextEvent(InputMethodContext& inputMethodContext, const InputMethodContext::EventData& inputMethodContextEvent);

  /**
   * @brief Callback when Clipboard signals an item should be pasted
   * @param[in] clipboard handle to Clipboard Event Notifier
   */
  void OnClipboardTextSelected(ClipboardEventNotifier& clipboard);

  /**
   * @brief Get a Property Map for the image used for the required Handle Image
   * @param[out] value the returned image property
   * @param[in] handleType the type of handle
   * @param[in] handleImageType the type of image for the given handleType
   */
  void GetHandleImagePropertyValue(Property::Value& value, Text::HandleType handleType, Text::HandleImageType handleImageType);

  /**
   * @brief Enable or disable clipping.
   */
  void EnableClipping();

  /**
   * @brief Callback when keyboard is shown/hidden.
   *
   * @param[in] keyboardShown True if keyboard is shown.
   */
  void KeyboardStatusChanged(bool keyboardShown);

  /**
   * @brief Callback when TextField is touched
   *
   * @param[in] actor TextField touched
   * @param[in] touch Touch information
   */
  bool OnTouched(Actor actor, const TouchEvent& touch);

  /**
   * @brief Callbacks called on idle.
   *
   * If there are notifications of change of input style on the queue, Toolkit::TextField::InputStyleChangedSignal() are emitted.
   */
  void OnIdleSignal();

  /**
   * @brief Emits TextChanged signal.
   */
  void EmitTextChangedSignal();

  /**
   * @brief Emits CursorPositionChanged signal.
   */
  void EmitCursorPositionChangedSignal();

  /**
   * @brief Emits SelectionChanged signal.
   */
  void EmitSelectionChangedSignal();

  /**
   * @brief Emits SelectionCleared signal.
   */
  void EmitSelectionClearedSignal();

  /**
   * @brief Callback function for when the layout is changed.
   * @param[in] actor The actor whose layoutDirection is changed.
   * @param[in] type  The layoutDirection.
   */
  void OnLayoutDirectionChanged(Actor actor, LayoutDirection::Type type);

  /**
   * Construct a new TextField.
   */
  TextField();

  /**
   * A reference counted object may only be deleted by calling Unreference()
   */
  virtual ~TextField();

  // Undefined copy constructor and assignment operators
  TextField(const TextField&);
  TextField& operator=(const TextField& rhs);

  /**
   * @brief Resize actor to the given size.
   *
   * @param[in] actor The actor to be resized.
   * @param[in] size Size to change.
   */
  void ResizeActor(Actor& actor, const Vector2& size);

  /**
   * @brief Render view, create and attach actor(s) to this Text Field.
   */
  void RenderText(Text::Controller::UpdateTextType updateTextType);

  // Connection needed to re-render text, when a Text Field returns to the scene.
  void OnSceneConnect(Dali::Actor actor);

  // Needed to synchronize TextAnchor actors with Anchor objects in text's logical model
  void OnAccessibilityStatusChanged();

private: // Data
  // Signals
  Toolkit::TextField::TextChangedSignalType                mTextChangedSignal;
  Toolkit::TextField::MaxLengthReachedSignalType           mMaxLengthReachedSignal;
  Toolkit::TextField::InputStyleChangedSignalType          mInputStyleChangedSignal;
  Toolkit::DevelTextField::AnchorClickedSignalType         mAnchorClickedSignal;
  Toolkit::DevelTextField::InputFilteredSignalType         mInputFilteredSignal;
  Toolkit::DevelTextField::CursorPositionChangedSignalType mCursorPositionChangedSignal;
  Toolkit::DevelTextField::SelectionChangedSignalType      mSelectionChangedSignal;
  Toolkit::DevelTextField::SelectionClearedSignalType      mSelectionClearedSignal;

  InputMethodContext               mInputMethodContext;
  Text::ControllerPtr              mController;
  Text::RendererPtr                mRenderer;
  Text::DecoratorPtr               mDecorator;
  Toolkit::Control                 mStencil;                  ///< For EXCEED_POLICY_CLIP
  std::vector<Actor>               mClippingDecorationActors; ///< Decoration actors which need clipping.
  std::vector<Toolkit::TextAnchor> mAnchorActors;
  Dali::InputMethodOptions         mInputMethodOptions;

  Actor         mRenderableActor;
  Actor         mActiveLayer;
  Actor         mBackgroundActor;
  CallbackBase* mIdleCallback;

  float mAlignmentOffset;
  int   mRenderingBackend;
  int   mExceedPolicy;
  bool  mHasBeenStaged : 1;
  bool  mTextChanged : 1;           ///< If true, emits TextChangedSignal in next OnRelayout().
  bool  mCursorPositionChanged : 1; ///< If true, emits CursorPositionChangedSignal at the end of OnRelayout().
  bool  mSelectionChanged : 1;      ///< If true, emits SelectionChangedSignal at the end of OnRelayout().
  bool  mSelectionCleared : 1;      ///< If true, emits SelectionClearedSignal at the end of OnRelayout().

  //args for cursor position changed event
  unsigned int mOldPosition;

  //args for selection changed event
  uint32_t mOldSelectionStart;
  uint32_t mOldSelectionEnd;

protected:
  struct PropertyHandler;

  /**
   * @brief This structure is to connect TextField with Accessible functions.
   */
  struct AccessibleImpl : public DevelControl::ControlAccessible,
                          public virtual Dali::Accessibility::EditableText,
                          public virtual Dali::Accessibility::Hypertext
  {
    using DevelControl::ControlAccessible::ControlAccessible;

    /**
     * @copydoc Dali::Accessibility::Accessible::GetName()
     */
    std::string GetName() const override;

    /**
     * @copydoc Dali::Accessibility::Text::GetText()
     */
    std::string GetText(size_t startOffset, size_t endOffset) const override;

    /**
     * @copydoc Dali::Accessibility::Text::GetCharacterCount()
     */
    size_t GetCharacterCount() const override;

    /**
     * @copydoc Dali::Accessibility::Text::GetCursorOffset()
     */
    size_t GetCursorOffset() const override;

    /**
     * @copydoc Dali::Accessibility::Text::SetCursorOffset()
     */
    bool SetCursorOffset(size_t offset) override;

    /**
     * @copydoc Dali::Accessibility::Text::GetTextAtOffset()
     */
    Accessibility::Range GetTextAtOffset(size_t offset, Accessibility::TextBoundary boundary) const override;

    /**
     * @copydoc Dali::Accessibility::Text::GetRangeOfSelection()
     */
    Accessibility::Range GetRangeOfSelection(size_t selectionIndex) const override;

    /**
     * @copydoc Dali::Accessibility::Text::RemoveSelection()
     */
    bool RemoveSelection(size_t selectionIndex) override;

    /**
     * @copydoc Dali::Accessibility::Text::SetRangeOfSelection()
     */
    bool SetRangeOfSelection(size_t selectionIndex, size_t startOffset, size_t endOffset) override;

    /**
     * @copydoc Dali::Accessibility::EditableText::CopyText()
     */
    bool CopyText(size_t startPosition, size_t endPosition) override;

    /**
     * @copydoc Dali::Accessibility::EditableText::CutText()
     */
    bool CutText(size_t startPosition, size_t endPosition) override;

    /**
     * @copydoc Dali::Accessibility::Accessible::GetStates()
     */
    Accessibility::States CalculateStates() override;

    /**
     * @copydoc Dali::Accessibility::EditableText::InsertText()
     */
    bool InsertText(size_t startPosition, std::string text) override;

    /**
     * @copydoc Dali::Accessibility::EditableText::SetTextContents()
     */
    bool SetTextContents(std::string newContents) override;

    /**
     * @copydoc Dali::Accessibility::EditableText::DeleteText()
     */
    bool DeleteText(size_t startPosition, size_t endPosition) override;

    /**
     * @copydoc Dali::Accessibility::Hypertext::GetLink()
     */
    Accessibility::Hyperlink* GetLink(int32_t linkIndex) const override;

    /**
     * @copydoc Dali::Accessibility::Hypertext::GetLinkIndex()
     */
    int32_t GetLinkIndex(int32_t characterOffset) const override;

    /**
     * @copydoc Dali::Accessibility::Hypertext::GetLinkCount()
     */
    int32_t GetLinkCount() const override;
  };
};

} // namespace Internal

// Helpers for public-api forwarding methods

inline Toolkit::Internal::TextField& GetImpl(Toolkit::TextField& textField)
{
  DALI_ASSERT_ALWAYS(textField);

  Dali::RefObject& handle = textField.GetImplementation();

  return static_cast<Toolkit::Internal::TextField&>(handle);
}

inline const Toolkit::Internal::TextField& GetImpl(const Toolkit::TextField& textField)
{
  DALI_ASSERT_ALWAYS(textField);

  const Dali::RefObject& handle = textField.GetImplementation();

  return static_cast<const Toolkit::Internal::TextField&>(handle);
}

} // namespace Toolkit

} // namespace Dali

#endif // DALI_TOOLKIT_INTERNAL_TEXT_FIELD_H
